from pathlib import Path

import pytest

from eegdash.api import EEGDashDataset
from eegdash.dataset.dataset import EEGChallengeDataset


def _dummy_record(dataset: str, ext: str = ".set") -> dict:
    # Minimal record used by EEGDashBaseDataset without triggering IO
    # bidspath must start with dataset root
    bidspath = f"{dataset}/sub-01/ses-01/eeg/sub-01_ses-01_task-test_run-01_eeg{ext}"
    return {
        "data_name": f"{dataset}_sub-01_ses-01_task-test_run-01_eeg{ext}",
        "dataset": dataset,
        "bidspath": bidspath,
        "bidsdependencies": [],
        # BIDS entities used to construct BIDSPath
        "subject": "01",
        "session": "01",
        "task": "test",
        "run": "01",
        # Not used in this test path, but present in real records
        "modality": "eeg",
        "sampling_frequency": 100.0,
        "nchans": 3,
        "ntimes": 100,
    }


@pytest.mark.parametrize(
    "release,dataset_id",
    [("R5", "ds005509")],
)
def test_dataset_folder_suffixes(cache_dir: Path, release: str, dataset_id: str):
    # Baseline EEGDashDataset should use plain dataset folder
    rec = _dummy_record(dataset_id)
    ds_plain = EEGDashDataset(cache_dir=cache_dir, records=[rec])
    base = ds_plain.datasets[0]
    assert base.bids_root == cache_dir / dataset_id
    assert str(base.filecache).startswith(str((cache_dir / dataset_id).resolve()))

    # EEGChallengeDataset mini=True should suffix with -bdf-mini
    ds_min = EEGChallengeDataset(release=release, cache_dir=cache_dir, records=[rec])
    base_min = ds_min.datasets[0]
    assert base_min.bids_root == cache_dir / f"{dataset_id}-bdf-mini"
    assert str(base_min.filecache).startswith(
        str((cache_dir / f"{dataset_id}-bdf-mini").resolve())
    )

    # EEGChallengeDataset mini=False should suffix with -bdf
    ds_full = EEGChallengeDataset(
        release=release, cache_dir=cache_dir, mini=False, records=[rec]
    )
    base_full = ds_full.datasets[0]
    assert base_full.bids_root == cache_dir / f"{dataset_id}-bdf"
    assert str(base_full.filecache).startswith(
        str((cache_dir / f"{dataset_id}-bdf").resolve())
    )



# Authors: The EEGDash contributors.
# License: GNU General Public License
# Copyright the EEGDash contributors.

"""Path utilities and cache directory management.

This module provides functions for resolving consistent cache directories and path
management throughout the EEGDash package, with integration to MNE-Python's
configuration system.
"""

from __future__ import annotations

import os
from pathlib import Path

from mne.utils import get_config as mne_get_config


def get_default_cache_dir() -> Path:
    """Resolve the default cache directory for EEGDash data.

    The function determines the cache directory based on the following
    priority order:
    1. The path specified by the ``EEGDASH_CACHE_DIR`` environment variable.
    2. The path specified by the ``MNE_DATA`` configuration in the MNE-Python
       config file.
    3. A hidden directory named ``.eegdash_cache`` in the current working
       directory.

    Returns
    -------
    pathlib.Path
        The resolved, absolute path to the default cache directory.

    """
    # 1) Explicit env var wins
    env_dir = os.environ.get("EEGDASH_CACHE_DIR")
    if env_dir:
        return Path(env_dir).expanduser().resolve()

    # 2) Reuse MNE's data cache location if configured
    mne_data = mne_get_config("MNE_DATA")
    if mne_data:
        return Path(mne_data).expanduser().resolve()

    # 3) Default to a project-local hidden folder
    return Path.cwd() / ".eegdash_cache"


__all__ = ["get_default_cache_dir"]



# Authors: The EEGDash contributors.
# License: GNU General Public License
# Copyright the EEGDash contributors.

"""Data utilities and dataset classes for EEG data handling.

This module provides core dataset classes for working with EEG data in the EEGDash ecosystem,
including classes for individual recordings and collections of datasets. It integrates with
braindecode for machine learning workflows and handles data loading from both local and remote sources.
"""

import io
import json
import os
import re
import traceback
from contextlib import redirect_stderr
from pathlib import Path
from typing import Any

import mne
import mne_bids
import numpy as np
import pandas as pd
from joblib import Parallel, delayed
from mne._fiff.utils import _read_segments_file
from mne.io import BaseRaw
from mne.utils.check import _soft_import
from mne_bids import BIDSPath

from braindecode.datasets import BaseDataset

from . import downloader
from .bids_eeg_metadata import enrich_from_participants
from .logging import logger
from .paths import get_default_cache_dir


class EEGDashBaseDataset(BaseDataset):
    """A single EEG recording dataset.

    Represents a single EEG recording, typically hosted on a remote server (like AWS S3)
    and cached locally upon first access. This class is a subclass of
    :class:`braindecode.datasets.BaseDataset` and can be used with braindecode's
    preprocessing and training pipelines.

    Parameters
    ----------
    record : dict
        A fully resolved metadata record for the data to load.
    cache_dir : str
        The local directory where the data will be cached.
    s3_bucket : str, optional
        The S3 bucket to download data from. If not provided, defaults to the
        OpenNeuro bucket.
    **kwargs
        Additional keyword arguments passed to the
        :class:`braindecode.datasets.BaseDataset` constructor.

    """

    _AWS_BUCKET = "s3://openneuro.org"

    def __init__(
        self,
        record: dict[str, Any],
        cache_dir: str,
        s3_bucket: str | None = None,
        **kwargs,
    ):
        super().__init__(None, **kwargs)
        self.record = record
        self.cache_dir = Path(cache_dir)
        self.bids_kwargs = self._get_raw_bids_args()

        if s3_bucket:
            self.s3_bucket = s3_bucket
            self.s3_open_neuro = False
        else:
            self.s3_bucket = self._AWS_BUCKET
            self.s3_open_neuro = True

        # Compute a dataset folder name under cache_dir that encodes preprocessing
        # (e.g., bdf, mini) to avoid overlapping with the original dataset cache.
        self.dataset_folder = record.get("dataset", "")
        # TODO: remove this hack when competition is over
        if s3_bucket:
            suffixes: list[str] = []
            bucket_lower = str(s3_bucket).lower()
            if "bdf" in bucket_lower:
                suffixes.append("bdf")
            if "mini" in bucket_lower:
                suffixes.append("mini")
            if suffixes:
                self.dataset_folder = f"{self.dataset_folder}-{'-'.join(suffixes)}"

        # Place files under the dataset-specific folder (with suffix if any)
        rel = Path(record["bidspath"])  # usually starts with dataset id
        if rel.parts and rel.parts[0] == record.get("dataset"):
            rel = Path(self.dataset_folder, *rel.parts[1:])
        else:
            rel = Path(self.dataset_folder) / rel
        self.filecache = self.cache_dir / rel
        self.bids_root = self.cache_dir / self.dataset_folder

        self.bidspath = BIDSPath(
            root=self.bids_root,
            datatype="eeg",
            suffix="eeg",
            **self.bids_kwargs,
        )

        self.s3file = downloader.get_s3path(self.s3_bucket, record["bidspath"])
        self.bids_dependencies = record["bidsdependencies"]
        self.bids_dependencies_original = record["bidsdependencies"]
        # TODO: removing temporary fix for BIDS dependencies path
        # when the competition is over and dataset is digested properly
        if not self.s3_open_neuro:
            self.bids_dependencies = [
                dep.split("/", 1)[1] for dep in self.bids_dependencies
            ]

        self._raw = None

    def _get_raw_bids_args(self) -> dict[str, Any]:
        """Extract BIDS-related arguments from the metadata record."""
        desired_fields = ["subject", "session", "task", "run"]
        return {k: self.record[k] for k in desired_fields if self.record[k]}

    def _ensure_raw(self) -> None:
        """Ensure the raw data file and its dependencies are cached locally."""
        # TO-DO: remove this once is fixed on the our side
        # for the competition
        if not self.s3_open_neuro:
            self.bidspath = self.bidspath.update(extension=".bdf")
            self.filecache = self.filecache.with_suffix(".bdf")

        if not os.path.exists(self.filecache):  # not preload
            if self.bids_dependencies:
                downloader.download_dependencies(
                    s3_bucket=self.s3_bucket,
                    bids_dependencies=self.bids_dependencies,
                    bids_dependencies_original=self.bids_dependencies_original,
                    cache_dir=self.cache_dir,
                    dataset_folder=self.dataset_folder,
                    record=self.record,
                    s3_open_neuro=self.s3_open_neuro,
                )
            self.filecache = downloader.download_s3_file(
                self.s3file, self.filecache, self.s3_open_neuro
            )
            self.filenames = [self.filecache]
        if self._raw is None:
            try:
                # mne-bids can emit noisy warnings to stderr; keep user logs clean
                _stderr_buffer = io.StringIO()
                with redirect_stderr(_stderr_buffer):
                    self._raw = mne_bids.read_raw_bids(
                        bids_path=self.bidspath, verbose="ERROR"
                    )
                # Enrich Raw.info and description with participants.tsv extras
                enrich_from_participants(
                    self.bids_root, self.bidspath, self._raw, self.description
                )

            except Exception as e:
                logger.error(
                    f"Error while reading BIDS file: {self.bidspath}\n"
                    "This may be due to a missing or corrupted file.\n"
                    "Please check the file and try again.\n"
                    "Usually erasing the local cache and re-downloading helps.\n"
                    f"`rm {self.bidspath}`"
                )
                logger.error(f"Exception: {e}")
                logger.error(traceback.format_exc())
                raise e

    def __len__(self) -> int:
        """Return the number of samples in the dataset."""
        if self._raw is None:
            if (
                self.record["ntimes"] is None
                or self.record["sampling_frequency"] is None
            ):
                self._ensure_raw()
            else:
                # FIXME: this is a bit strange and should definitely not change as a side effect
                #  of accessing the data (which it will, since ntimes is the actual length but rounded down)
                return int(self.record["ntimes"] * self.record["sampling_frequency"])
        return len(self._raw)

    @property
    def raw(self) -> BaseRaw:
        """The MNE Raw object for this recording.

        Accessing this property triggers the download and caching of the data
        if it has not been accessed before.

        Returns
        -------
        mne.io.BaseRaw
            The loaded MNE Raw object.

        """
        if self._raw is None:
            self._ensure_raw()
        return self._raw

    @raw.setter
    def raw(self, raw: BaseRaw):
        self._raw = raw


class EEGDashBaseRaw(BaseRaw):
    """MNE BaseRaw wrapper for automatic S3 data fetching.

    This class extends :class:`mne.io.BaseRaw` to automatically fetch data
    from an S3 bucket and cache it locally when data is first accessed.
    It is intended for internal use within the EEGDash ecosystem.

    Parameters
    ----------
    input_fname : str
        The path to the file on the S3 bucket (relative to the bucket root).
    metadata : dict
        The metadata record for the recording, containing information like
        sampling frequency, channel names, etc.
    preload : bool, default False
        If True, preload the data into memory.
    cache_dir : str, optional
        Local directory for caching data. If None, a default directory is used.
    bids_dependencies : list of str, default []
        A list of BIDS metadata files to download alongside the main recording.
    verbose : str, int, or None, default None
        The MNE verbosity level.

    See Also
    --------
    mne.io.Raw : The base class for Raw objects in MNE.

    """

    _AWS_BUCKET = "s3://openneuro.org"

    def __init__(
        self,
        input_fname: str,
        metadata: dict[str, Any],
        preload: bool = False,
        *,
        cache_dir: str | None = None,
        bids_dependencies: list[str] = [],
        verbose: Any = None,
    ):
        # Create a simple RawArray
        sfreq = metadata["sfreq"]  # Sampling frequency
        n_times = metadata["n_times"]
        ch_names = metadata["ch_names"]
        ch_types = []
        for ch in metadata["ch_types"]:
            chtype = ch.lower()
            if chtype == "heog" or chtype == "veog":
                chtype = "eog"
            ch_types.append(chtype)
        info = mne.create_info(ch_names=ch_names, sfreq=sfreq, ch_types=ch_types)

        self.s3file = downloader.get_s3path(self._AWS_BUCKET, input_fname)
        self.cache_dir = Path(cache_dir) if cache_dir else get_default_cache_dir()
        self.filecache = self.cache_dir / input_fname
        self.bids_dependencies = bids_dependencies

        if preload and not os.path.exists(self.filecache):
            self.filecache = downloader.download_s3_file(
                self.s3file, self.filecache, self.s3_open_neuro
            )
            self.filenames = [self.filecache]
            preload = self.filecache

        super().__init__(
            info,
            preload,
            last_samps=[n_times - 1],
            orig_format="single",
            verbose=verbose,
        )

    def _read_segment(
        self, start=0, stop=None, sel=None, data_buffer=None, *, verbose=None
    ):
        """Read a segment of data, downloading if necessary."""
        if not os.path.exists(self.filecache):  # not preload
            if self.bids_dependencies:  # this is use only to sidecars for now
                downloader.download_dependencies(
                    s3_bucket=self._AWS_BUCKET,
                    bids_dependencies=self.bids_dependencies,
                    bids_dependencies_original=None,
                    cache_dir=self.cache_dir,
                    dataset_folder=self.filecache,
                    record={},
                    s3_open_neuro=self.s3_open_neuro,
                )
            self.filecache = downloader.download_s3_file(
                self.s3file, self.filecache, self.s3_open_neuro
            )
            self.filenames = [self.filecache]
        else:  # not preload and file is not cached
            self.filenames = [self.filecache]
        return super()._read_segment(start, stop, sel, data_buffer, verbose=verbose)

    def _read_segment_file(self, data, idx, fi, start, stop, cals, mult):
        """Read a chunk of data from a local file."""
        _read_segments_file(self, data, idx, fi, start, stop, cals, mult, dtype="<f4")


class EEGBIDSDataset:
    """An interface to a local BIDS dataset containing EEG recordings.

    This class centralizes interactions with a BIDS dataset on the local
    filesystem, providing methods to parse metadata, find files, and
    retrieve BIDS-related information.

    Parameters
    ----------
    data_dir : str or Path
        The path to the local BIDS dataset directory.
    dataset : str
        A name for the dataset (e.g., "ds002718").

    """

    ALLOWED_FILE_FORMAT = ["eeglab", "brainvision", "biosemi", "european"]
    RAW_EXTENSIONS = {
        ".set": [".set", ".fdt"],  # eeglab
        ".edf": [".edf"],  # european
        ".vhdr": [".eeg", ".vhdr", ".vmrk", ".dat", ".raw"],  # brainvision
        ".bdf": [".bdf"],  # biosemi
    }
    METADATA_FILE_EXTENSIONS = [
        "eeg.json",
        "channels.tsv",
        "electrodes.tsv",
        "events.tsv",
        "events.json",
    ]

    def __init__(
        self,
        data_dir=None,  # location of bids dataset
        dataset="",  # dataset name
    ):
        bids_lib = _soft_import("bids", purpose="digestion of datasets", strict=False)

        if bids_lib is None:
            raise ImportError(
                "The 'pybids' package is required to use EEGBIDSDataset. "
                "Please install it via 'pip install eegdash[digestion]'."
            )

        if data_dir is None or not os.path.exists(data_dir):
            raise ValueError("data_dir must be specified and must exist")
        self.bidsdir = Path(data_dir)
        self.dataset = dataset
        # Accept exact dataset folder or a variant with informative suffixes
        # (e.g., dsXXXXX-bdf, dsXXXXX-bdf-mini) to avoid collisions.
        dir_name = self.bidsdir.name
        if not (dir_name == self.dataset or dir_name.startswith(self.dataset + "-")):
            raise AssertionError(
                f"BIDS directory '{dir_name}' does not correspond to dataset '{self.dataset}'"
            )
        self.layout = bids_lib.BIDSLayout(data_dir)

        # get all recording files in the bids directory
        self.files = self._get_recordings(self.layout)
        assert len(self.files) > 0, ValueError(
            "Unable to construct EEG dataset. No EEG recordings found."
        )
        assert self.check_eeg_dataset(), ValueError("Dataset is not an EEG dataset.")

    def check_eeg_dataset(self) -> bool:
        """Check if the BIDS dataset contains EEG data.

        Returns
        -------
        bool
            True if the dataset's modality is EEG, False otherwise.

        """
        return self.get_bids_file_attribute("modality", self.files[0]).lower() == "eeg"

    def _get_recordings(self, layout) -> list[str]:
        """Get a list of all EEG recording files in the BIDS layout."""
        files = []
        for ext, exts in self.RAW_EXTENSIONS.items():
            files = layout.get(extension=ext, return_type="filename")
            if files:
                break
        return files

    def _get_relative_bidspath(self, filename: str) -> str:
        """Make a file path relative to the BIDS parent directory."""
        bids_parent_dir = self.bidsdir.parent.absolute()
        return str(Path(filename).relative_to(bids_parent_dir))

    def _get_property_from_filename(self, property: str, filename: str) -> str:
        """Parse a BIDS entity from a filename."""
        import platform

        if platform.system() == "Windows":
            lookup = re.search(rf"{property}-(.*?)[_\\]", filename)
        else:
            lookup = re.search(rf"{property}-(.*?)[_\/]", filename)
        return lookup.group(1) if lookup else ""

    def _merge_json_inheritance(self, json_files: list[str | Path]) -> dict:
        """Merge a list of JSON files according to BIDS inheritance."""
        json_files.reverse()
        json_dict = {}
        for f in json_files:
            with open(f) as fp:
                json_dict.update(json.load(fp))
        return json_dict

    def _get_bids_file_inheritance(
        self, path: str | Path, basename: str, extension: str
    ) -> list[Path]:
        """Find all applicable metadata files using BIDS inheritance."""
        top_level_files = ["README", "dataset_description.json", "participants.tsv"]
        bids_files = []

        if isinstance(path, str):
            path = Path(path)
        if not path.exists():
            raise ValueError(f"path {path} does not exist")

        for file in os.listdir(path):
            if os.path.isfile(path / file) and file.endswith(extension):
                bids_files.append(path / file)

        if any(file in os.listdir(path) for file in top_level_files):
            return bids_files
        else:
            bids_files.extend(
                self._get_bids_file_inheritance(path.parent, basename, extension)
            )
            return bids_files

    def get_bids_metadata_files(
        self, filepath: str | Path, metadata_file_extension: str
    ) -> list[Path]:
        """Retrieve all metadata files that apply to a given data file.

        Follows the BIDS inheritance principle to find all relevant metadata
        files (e.g., ``channels.tsv``, ``eeg.json``) for a specific recording.

        Parameters
        ----------
        filepath : str or Path
            The path to the data file.
        metadata_file_extension : str
            The extension of the metadata file to search for (e.g., "channels.tsv").

        Returns
        -------
        list of Path
            A list of paths to the matching metadata files.

        """
        if isinstance(filepath, str):
            filepath = Path(filepath)
        if not filepath.exists():
            raise ValueError(f"filepath {filepath} does not exist")
        path, filename = os.path.split(filepath)
        basename = filename[: filename.rfind("_")]
        meta_files = self._get_bids_file_inheritance(
            path, basename, metadata_file_extension
        )
        return meta_files

    def _scan_directory(self, directory: str, extension: str) -> list[Path]:
        """Scan a directory for files with a given extension."""
        result_files = []
        directory_to_ignore = [".git", ".datalad", "derivatives", "code"]
        with os.scandir(directory) as entries:
            for entry in entries:
                if entry.is_file() and entry.name.endswith(extension):
                    result_files.append(Path(entry.path))
                elif entry.is_dir() and not any(
                    name in entry.name for name in directory_to_ignore
                ):
                    result_files.append(Path(entry.path))
        return result_files

    def _get_files_with_extension_parallel(
        self, directory: str, extension: str = ".set", max_workers: int = -1
    ) -> list[Path]:
        """Scan a directory tree in parallel for files with a given extension."""
        result_files = []
        dirs_to_scan = [directory]

        while dirs_to_scan:
            logger.info(
                f"Directories to scan: {len(dirs_to_scan)}, files: {dirs_to_scan}"
            )
            results = Parallel(n_jobs=max_workers, prefer="threads", verbose=1)(
                delayed(self._scan_directory)(d, extension) for d in dirs_to_scan
            )

            dirs_to_scan = []
            for res in results:
                for path in res:
                    if os.path.isdir(path):
                        dirs_to_scan.append(path)
                    else:
                        result_files.append(path)
            logger.info(f"Found {len(result_files)} files.")

        return result_files

    def load_and_preprocess_raw(
        self, raw_file: str, preprocess: bool = False
    ) -> np.ndarray:
        """Load and optionally preprocess a raw data file.

        This is a utility function for testing or debugging, not for general use.

        Parameters
        ----------
        raw_file : str
            Path to the raw EEGLAB file (.set).
        preprocess : bool, default False
            If True, apply a high-pass filter, notch filter, and resample the data.

        Returns
        -------
        numpy.ndarray
            The loaded and processed data as a NumPy array.

        """
        logger.info(f"Loading raw data from {raw_file}")
        EEG = mne.io.read_raw_eeglab(raw_file, preload=True, verbose="error")

        if preprocess:
            EEG = EEG.filter(l_freq=0.25, h_freq=25, verbose=False)
            EEG = EEG.notch_filter(freqs=(60), verbose=False)
            sfreq = 128
            if EEG.info["sfreq"] != sfreq:
                EEG = EEG.resample(sfreq)

        mat_data = EEG.get_data()

        if len(mat_data.shape) > 2:
            raise ValueError("Expect raw data to be CxT dimension")
        return mat_data

    def get_files(self) -> list[str]:
        """Get all EEG recording file paths in the BIDS dataset.

        Returns
        -------
        list of str
            A list of file paths for all valid EEG recordings.

        """
        return self.files

    def resolve_bids_json(self, json_files: list[str]) -> dict:
        """Resolve BIDS JSON inheritance and merge files.

        Parameters
        ----------
        json_files : list of str
            A list of JSON file paths, ordered from the lowest (most specific)
            to highest level of the BIDS hierarchy.

        Returns
        -------
        dict
            A dictionary containing the merged JSON data.

        """
        if not json_files:
            raise ValueError("No JSON files provided")
        json_files.reverse()

        json_dict = {}
        for json_file in json_files:
            with open(json_file) as f:
                json_dict.update(json.load(f))
        return json_dict

    def get_bids_file_attribute(self, attribute: str, data_filepath: str) -> Any:
        """Retrieve a specific attribute from BIDS metadata.

        Parameters
        ----------
        attribute : str
            The name of the attribute to retrieve (e.g., "sfreq", "subject").
        data_filepath : str
            The path to the data file.

        Returns
        -------
        Any
            The value of the requested attribute, or None if not found.

        """
        entities = self.layout.parse_file_entities(data_filepath)
        bidsfile = self.layout.get(**entities)[0]
        attributes = bidsfile.get_entities(metadata="all")
        attribute_mapping = {
            "sfreq": "SamplingFrequency",
            "modality": "datatype",
            "task": "task",
            "session": "session",
            "run": "run",
            "subject": "subject",
            "ntimes": "RecordingDuration",
            "nchans": "EEGChannelCount",
        }
        attribute_value = attributes.get(attribute_mapping.get(attribute), None)
        return attribute_value

    def channel_labels(self, data_filepath: str) -> list[str]:
        """Get a list of channel labels from channels.tsv.

        Parameters
        ----------
        data_filepath : str
            The path to the data file.

        Returns
        -------
        list of str
            A list of channel names.

        """
        channels_tsv = pd.read_csv(
            self.get_bids_metadata_files(data_filepath, "channels.tsv")[0], sep="\t"
        )
        return channels_tsv["name"].tolist()

    def channel_types(self, data_filepath: str) -> list[str]:
        """Get a list of channel types from channels.tsv.

        Parameters
        ----------
        data_filepath : str
            The path to the data file.

        Returns
        -------
        list of str
            A list of channel types.

        """
        channels_tsv = pd.read_csv(
            self.get_bids_metadata_files(data_filepath, "channels.tsv")[0], sep="\t"
        )
        return channels_tsv["type"].tolist()

    def num_times(self, data_filepath: str) -> int:
        """Get the number of time points in the recording.

        Calculated from ``SamplingFrequency`` and ``RecordingDuration`` in eeg.json.

        Parameters
        ----------
        data_filepath : str
            The path to the data file.

        Returns
        -------
        int
            The approximate number of time points.

        """
        eeg_jsons = self.get_bids_metadata_files(data_filepath, "eeg.json")
        eeg_json_dict = self._merge_json_inheritance(eeg_jsons)
        return int(
            eeg_json_dict["SamplingFrequency"] * eeg_json_dict["RecordingDuration"]
        )

    def subject_participant_tsv(self, data_filepath: str) -> dict[str, Any]:
        """Get the participants.tsv record for a subject.

        Parameters
        ----------
        data_filepath : str
            The path to a data file belonging to the subject.

        Returns
        -------
        dict
            A dictionary of the subject's information from participants.tsv.

        """
        participants_tsv_path = self.get_bids_metadata_files(
            data_filepath, "participants.tsv"
        )[0]
        participants_tsv = pd.read_csv(participants_tsv_path, sep="\t")
        if participants_tsv.empty:
            return {}
        participants_tsv.set_index("participant_id", inplace=True)
        subject = f"sub-{self.get_bids_file_attribute('subject', data_filepath)}"
        return participants_tsv.loc[subject].to_dict()

    def eeg_json(self, data_filepath: str) -> dict[str, Any]:
        """Get the merged eeg.json metadata for a data file.

        Parameters
        ----------
        data_filepath : str
            The path to the data file.

        Returns
        -------
        dict
            The merged eeg.json metadata.

        """
        eeg_jsons = self.get_bids_metadata_files(data_filepath, "eeg.json")
        return self._merge_json_inheritance(eeg_jsons)

    def channel_tsv(self, data_filepath: str) -> dict[str, Any]:
        """Get the channels.tsv metadata as a dictionary.

        Parameters
        ----------
        data_filepath : str
            The path to the data file.

        Returns
        -------
        dict
            The channels.tsv data, with columns as keys.

        """
        channels_tsv_path = self.get_bids_metadata_files(data_filepath, "channels.tsv")[
            0
        ]
        channels_tsv = pd.read_csv(channels_tsv_path, sep="\t")
        channel_tsv_dict = channels_tsv.to_dict()
        for list_field in ["name", "type", "units"]:
            if list_field in channel_tsv_dict:
                channel_tsv_dict[list_field] = list(
                    channel_tsv_dict[list_field].values()
                )
        return channel_tsv_dict


__all__ = ["EEGDashBaseDataset", "EEGBIDSDataset", "EEGDashBaseRaw"]



import copy
from collections.abc import Callable
from typing import Dict, List

import numpy as np
import pandas as pd
from joblib import Parallel, delayed
from torch.utils.data import DataLoader
from tqdm import tqdm

from braindecode.datasets.base import (
    BaseConcatDataset,
    EEGWindowsDataset,
    WindowsDataset,
)

from .datasets import FeaturesConcatDataset, FeaturesDataset
from .extractors import FeatureExtractor


def _extract_features_from_windowsdataset(
    win_ds: EEGWindowsDataset | WindowsDataset,
    feature_extractor: FeatureExtractor,
    batch_size: int = 512,
) -> FeaturesDataset:
    """Extract features from a single `WindowsDataset`.

    This is a helper function that iterates through a `WindowsDataset` in
    batches, applies a `FeatureExtractor`, and returns the results as a
    `FeaturesDataset`.

    Parameters
    ----------
    win_ds : EEGWindowsDataset or WindowsDataset
        The windowed dataset to extract features from.
    feature_extractor : FeatureExtractor
        The feature extractor instance to apply.
    batch_size : int, default 512
        The number of windows to process in each batch.

    Returns
    -------
    FeaturesDataset
        A new dataset containing the extracted features and associated metadata.

    """
    metadata = win_ds.metadata
    if not win_ds.targets_from == "metadata":
        metadata = copy.deepcopy(metadata)
        metadata["orig_index"] = metadata.index
        metadata.set_index(
            ["i_window_in_trial", "i_start_in_trial", "i_stop_in_trial"],
            drop=False,
            inplace=True,
        )
    win_dl = DataLoader(win_ds, batch_size=batch_size, shuffle=False, drop_last=False)
    features_dict = dict()
    ch_names = win_ds.raw.ch_names
    for X, y, crop_inds in win_dl:
        X = X.numpy()
        if hasattr(y, "tolist"):
            y = y.tolist()
        win_dict = dict()
        win_dict.update(
            feature_extractor(X, _batch_size=X.shape[0], _ch_names=ch_names)
        )
        if not win_ds.targets_from == "metadata":
            metadata.loc[crop_inds, "target"] = y
        for k, v in win_dict.items():
            if k not in features_dict:
                features_dict[k] = []
            features_dict[k].extend(v)
    features_df = pd.DataFrame(features_dict)
    if not win_ds.targets_from == "metadata":
        metadata.reset_index(drop=True, inplace=True)
        metadata.drop("orig_index", axis=1, inplace=True, errors="ignore")

    return FeaturesDataset(
        features_df,
        metadata=metadata,
        description=win_ds.description,
        raw_info=win_ds.raw.info,
        raw_preproc_kwargs=getattr(win_ds, "raw_preproc_kwargs", None),
        window_kwargs=getattr(win_ds, "window_kwargs", None),
        features_kwargs=feature_extractor.features_kwargs,
    )


def extract_features(
    concat_dataset: BaseConcatDataset,
    features: FeatureExtractor | Dict[str, Callable] | List[Callable],
    *,
    batch_size: int = 512,
    n_jobs: int = 1,
) -> FeaturesConcatDataset:
    """Extract features from a concatenated dataset of windows.

    This function applies a feature extractor to each `WindowsDataset` within a
    `BaseConcatDataset` in parallel and returns a `FeaturesConcatDataset`
    with the results.

    Parameters
    ----------
    concat_dataset : BaseConcatDataset
        A concatenated dataset of `WindowsDataset` or `EEGWindowsDataset`
        instances.
    features : FeatureExtractor or dict or list
        The feature extractor(s) to apply. Can be a `FeatureExtractor`
        instance, a dictionary of named feature functions, or a list of
        feature functions.
    batch_size : int, default 512
        The size of batches to use for feature extraction.
    n_jobs : int, default 1
        The number of parallel jobs to use for extracting features from the
        datasets.

    Returns
    -------
    FeaturesConcatDataset
        A new concatenated dataset containing the extracted features.

    """
    if isinstance(features, list):
        features = dict(enumerate(features))
    if not isinstance(features, FeatureExtractor):
        features = FeatureExtractor(features)
    feature_ds_list = list(
        tqdm(
            Parallel(n_jobs=n_jobs, return_as="generator")(
                delayed(_extract_features_from_windowsdataset)(
                    win_ds, features, batch_size
                )
                for win_ds in concat_dataset.datasets
            ),
            total=len(concat_dataset.datasets),
            desc="Extracting features",
        )
    )
    return FeaturesConcatDataset(feature_ds_list)


def fit_feature_extractors(
    concat_dataset: BaseConcatDataset,
    features: FeatureExtractor | Dict[str, Callable] | List[Callable],
    batch_size: int = 8192,
) -> FeatureExtractor:
    """Fit trainable feature extractors on a dataset.

    If the provided feature extractor (or any of its sub-extractors) is
    trainable (i.e., subclasses `TrainableFeature`), this function iterates
    through the dataset to fit it.

    Parameters
    ----------
    concat_dataset : BaseConcatDataset
        The dataset to use for fitting the feature extractors.
    features : FeatureExtractor or dict or list
        The feature extractor(s) to fit.
    batch_size : int, default 8192
        The batch size to use when iterating through the dataset for fitting.

    Returns
    -------
    FeatureExtractor
        The fitted feature extractor.

    """
    if isinstance(features, list):
        features = dict(enumerate(features))
    if not isinstance(features, FeatureExtractor):
        features = FeatureExtractor(features)
    if not features._is_trainable:
        return features
    features.clear()
    concat_dl = DataLoader(
        concat_dataset, batch_size=batch_size, shuffle=False, drop_last=False
    )
    for X, y, _ in tqdm(
        concat_dl, total=len(concat_dl), desc="Fitting feature extractors"
    ):
        features.partial_fit(X.numpy(), y=np.array(y))
    features.fit()
    return features



import csv
import importlib
import inspect
import os
import sys
from collections import Counter
from datetime import datetime, timezone
from pathlib import Path
from typing import Iterable, Mapping, Sequence

from sphinx.util import logging
from sphinx_gallery.sorting import ExplicitOrder, FileNameSortKey
from tabulate import tabulate

sys.path.insert(0, os.path.abspath(".."))

import eegdash

# -- Project information -----------------------------------------------------

project = "EEG Dash"
copyright = f"2025–{datetime.now(tz=timezone.utc).year}, {project} Developers"
author = "Bruno Aristimunha and Arnaud Delorme"
release = eegdash.__version__
version = ".".join(release.split(".")[:2])


# -- General configuration ---------------------------------------------------

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.doctest",
    "sphinx.ext.todo",
    "sphinx.ext.coverage",
    "sphinx.ext.mathjax",
    "sphinx.ext.ifconfig",
    "sphinx.ext.intersphinx",
    "sphinx.ext.githubpages",
    "sphinx.ext.linkcode",
    "sphinx.ext.graphviz",
    "sphinx.ext.napoleon",
    "sphinx_design",
    "numpydoc",
    "sphinx_gallery.gen_gallery",
    "sphinx_sitemap",
    "sphinx_copybutton",
    "sphinx.ext.graphviz",
    "sphinx_time_estimation",
]

templates_path = ["_templates"]
exclude_patterns = ["build", "Thumbs.db", ".DS_Store"]

# Autosummary: generate stub pages for documented items
autosummary_generate = True
# Include members that are imported into modules (e.g., re-exported dataset classes)
autosummary_imported_members = True
autosummary_ignore_module_all = False

# Suppress benign warnings
suppress_warnings = [
    # Sphinx-Gallery uses functions/classes in config which are not picklable
    "config.cache",
]

# -- Options for HTML output -------------------------------------------------

html_theme = "pydata_sphinx_theme"
html_static_path = ["_static"]
html_logo = "_static/eegdash_icon.svg"
html_favicon = "_static/eegdash_icon.png"
html_title = "EEG Dash"
html_short_title = "EEG Dash"
html_css_files = [
    "https://cdn.datatables.net/v/bm/dt-1.13.4/datatables.min.css",
    "https://cdn.datatables.net/buttons/2.4.2/css/buttons.dataTables.min.css",
    "https://cdn.datatables.net/select/1.7.0/css/select.dataTables.min.css",
    "https://cdn.datatables.net/searchpanes/2.3.1/css/searchPanes.dataTables.min.css",
    "custom.css",
]
html_js_files = [
    "https://code.jquery.com/jquery-3.7.1.min.js",
    "https://cdn.datatables.net/v/bm/dt-1.13.4/datatables.min.js",
    "https://cdn.datatables.net/buttons/2.4.2/js/dataTables.buttons.min.js",
    "https://cdn.datatables.net/select/1.7.0/js/dataTables.select.min.js",
    "https://cdn.datatables.net/searchpanes/2.3.1/js/dataTables.searchPanes.min.js",
    "js/tag-palette.js",
    "js/datatables-init.js",
]

# Required for sphinx-sitemap: set the canonical base URL of the site
# Make sure this matches the actual published docs URL and ends with '/'
html_baseurl = "https://eegdash.org/"

html_theme_options = {
    "icon_links_label": "External Links",  # for screen reader
    # Show an "Edit this page" button linking to GitHub
    "use_edit_page_button": True,
    "navigation_with_keys": False,
    "collapse_navigation": False,
    "header_links_before_dropdown": 6,
    "navigation_depth": 6,
    "show_nav_level": 2,
    "show_toc_level": 1,
    "navbar_end": ["theme-switcher", "navbar-icon-links"],
    "footer_start": ["copyright"],
    "logo": {
        "image_light": "_static/eegdash_long_white.svg",
        "image_dark": "_static/eegdash_long_dark.svg",
        "alt_text": "EEG Dash Logo",
    },
    "external_links": [
        {"name": "EEG2025", "url": "https://eeg2025.github.io/"},
    ],
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/sccn/EEGDash",
            "icon": "fa-brands fa-github",
            "type": "fontawesome",
        },
        {
            "name": "PyPI",
            "url": "https://pypi.org/project/eegdash/",
            "icon": "fa-solid fa-box",
            "type": "fontawesome",
        },
        {
            "name": "Docs (Stable)",
            "url": "https://eegdash.org/EEGDash",
            "icon": "fa-solid fa-book",
            "type": "fontawesome",
        },
        {
            "name": "Discord",
            "url": "https://discord.gg/8jd7nVKwsc",
            "icon": "fa-brands fa-discord",
            "type": "fontawesome",
        },
    ],
}

html_sidebars = {"api": [], "dataset_summary": [], "installation": []}

# Copy extra files (e.g., robots.txt) to the output root
html_extra_path = ["_extra"]

# Provide GitHub context so the edit button and custom templates
# (e.g., Sphinx-Gallery "Open in Colab") know where the source lives.
# These values should match the repository and docs location.
html_context = {
    "github_user": "sccn",
    "github_repo": "EEGDash",
    # Branch used to build and host the docs
    "github_version": "main",
    # Path to the documentation root within the repo
    "doc_path": "docs/source",
}


# Linkcode configuration: map documented objects to GitHub source lines
def _linkcode_resolve_py_domain(info):
    modname = info.get("module")
    fullname = info.get("fullname")
    if not modname:
        return None

    try:
        submod = sys.modules.get(modname)
        if submod is None:
            submod = importlib.import_module(modname)

        obj = submod
        for part in fullname.split("."):
            obj = getattr(obj, part)

        # Unwrap decorators to reach the actual implementation
        obj = inspect.unwrap(obj)
        fn = inspect.getsourcefile(obj) or inspect.getfile(obj)
        if not fn:
            return None
        fn = os.path.realpath(fn)

        # Compute line numbers
        try:
            source, start = inspect.getsourcelines(obj)
            end = start + len(source) - 1
            linespec = f"#L{start}-L{end}"
        except OSError:
            linespec = ""

        # Make path relative to repo root (parent of the installed package dir)
        pkg_dir = os.path.realpath(os.path.dirname(eegdash.__file__))
        repo_root = os.path.realpath(os.path.join(pkg_dir, os.pardir))
        rel_path = os.path.relpath(fn, start=repo_root)

        # Choose commit/branch for links; override via env if provided
        commit = os.environ.get(
            "LINKCODE_COMMIT", html_context.get("github_version", "main")
        )
        return f"https://github.com/{html_context['github_user']}/{html_context['github_repo']}/blob/{commit}/{rel_path}{linespec}"
    except Exception:
        return None


def linkcode_resolve(domain, info):
    if domain == "py":
        return _linkcode_resolve_py_domain(info)
    return None


# -- Extension configurations ------------------------------------------------
autoclass_content = "both"

# Numpydoc
numpydoc_show_class_members = False

# Sphinx Gallery
EX_DIR = "../../examples"  # relative to docs/source
sphinx_gallery_conf = {
    "examples_dirs": [f"{EX_DIR}"],
    "gallery_dirs": ["generated/auto_examples"],
    "nested_sections": False,
    "backreferences_dir": "gen_modules/backreferences",
    "inspect_global_variables": True,
    "show_memory": True,
    "show_api_usage": False,
    "doc_module": ("eegdash", "numpy", "scipy", "matplotlib"),
    "reference_url": {"eegdash": None},
    "filename_pattern": r"/(?:plot|tutorial)_(?!_).*\.py",
    "matplotlib_animations": True,
    "first_notebook_cell": (
        "# For tips on running notebooks in Google Colab:\n"
        "# `pip install eegdash`\n"
        "%matplotlib inline"
    ),
    "subsection_order": ExplicitOrder([f"{EX_DIR}/core", "*"]),
    "within_subsection_order": FileNameSortKey,
}

# -- Custom Setup Function to fix the error -----------------------------------


LOGGER = logging.getLogger(__name__)


AUTOGEN_NOTICE = """..
   This file is auto-generated during the Sphinx build.
   Do not edit by hand; changes will be overwritten.

"""


DATASET_PAGE_TEMPLATE = """{notice}{title}
{underline}

.. currentmodule:: eegdash.dataset

.. autoclass:: eegdash.dataset.{class_name}
   :members:
   :undoc-members:
   :show-inheritance:
   :inherited-members:
   :member-order: bysource

Dataset Information
-------------------

{metadata_section}

Usage Example
-------------

{usage_section}

See Also
--------

{see_also_section}

"""


DATASET_INDEX_TEMPLATE = """{notice}Datasets API
=======================

The :mod:`eegdash.dataset` package exposes dynamically registered dataset
classes. See :doc:`eegdash.dataset` for the module-level API, including
:class:`~eegdash.dataset.EEGChallengeDataset` and helper utilities.

Dataset Overview
----------------

EEGDash currently exposes **{dataset_count} OpenNeuro EEG datasets** that are
registered dynamically from mongo database. The table below summarises
the distribution by experimental type as tracked in the summary file.

Base Dataset API
----------------

.. toctree::
   :maxdepth: 1

   eegdash.dataset.EEGChallengeDataset

.. list-table:: Dataset counts by experimental type
   :widths: 60 20
   :header-rows: 1

   * - Experimental Type
     - Datasets
{experiment_rows}


All Datasets
------------

.. toctree::
   :maxdepth: 1
   :caption: Individual Datasets

{toctree_entries}

"""


BASE_DATASET_TEMPLATE = """{notice}.. _api_eegdash_challenge_dataset:

.. currentmodule:: eegdash.dataset

EEGChallengeDataset
===================

.. autoclass:: eegdash.dataset.EEGChallengeDataset
   :members:
   :undoc-members:
   :show-inheritance:
   :member-order: bysource

"""


def _write_if_changed(path: Path, content: str) -> bool:
    """Write ``content`` to ``path`` if it differs from the current file."""
    existing = path.read_text(encoding="utf-8") if path.exists() else None
    if existing == content:
        return False
    path.write_text(content, encoding="utf-8")
    return True


def _iter_dataset_classes() -> Sequence[str]:
    """Return the sorted dataset class names exported by ``eegdash.dataset``."""
    import eegdash.dataset as dataset_module  # local import for clarity
    from eegdash.api import EEGDashDataset

    class_names: list[str] = []
    for name in getattr(dataset_module, "__all__", []):
        if name == "EEGChallengeDataset":
            continue
        obj = getattr(dataset_module, name, None)
        if not inspect.isclass(obj):
            continue
        if not issubclass(obj, EEGDashDataset):
            continue
        if getattr(obj, "_dataset", None) is None:
            continue
        class_names.append(name)

    return tuple(sorted(class_names))


def _load_experiment_counts(dataset_names: Iterable[str]) -> list[tuple[str, int]]:
    """Return a sorted list of (experiment_type, count) pairs."""
    csv_path = Path(importlib.import_module("eegdash.dataset").__file__).with_name(
        "dataset_summary.csv"
    )
    if not csv_path.exists():
        return []

    counter: Counter[str] = Counter()
    valid_names = {name.upper() for name in dataset_names}

    with csv_path.open(encoding="utf-8") as handle:
        filtered = (
            line
            for line in handle
            if line.strip() and not line.lstrip().startswith("#")
        )
        reader = csv.DictReader(filtered)
        for row in reader:
            dataset_id = (row.get("dataset") or "").strip().upper()
            if dataset_id not in valid_names:
                continue
            exp_type = (row.get("type of exp") or "Unspecified").strip()
            counter[exp_type or "Unspecified"] += 1

    # Order by decreasing count then alphabetically for stable output
    return sorted(counter.items(), key=lambda item: (-item[1], item[0]))


def _render_experiment_rows(pairs: Iterable[tuple[str, int]]) -> str:
    lines = []
    for exp_type, count in pairs:
        label = exp_type or "Unspecified"
        lines.append(f"   * - {label}\n     - {count}")
    if not lines:
        lines.append("   * - No experimental metadata available\n     - N/A")
    return "\n".join(lines)


def _render_toctree_entries(names: Sequence[str]) -> str:
    return "\n".join(f"   eegdash.dataset.{name}" for name in names)


def _load_dataset_rows(dataset_names: Sequence[str]) -> Mapping[str, Mapping[str, str]]:
    csv_path = Path(importlib.import_module("eegdash.dataset").__file__).with_name(
        "dataset_summary.csv"
    )
    if not csv_path.exists():
        return {}

    wanted = set(dataset_names)
    rows: dict[str, Mapping[str, str]] = {}

    with csv_path.open(encoding="utf-8") as handle:
        filtered = (
            line
            for line in handle
            if line.strip() and not line.lstrip().startswith("#")
        )
        reader = csv.DictReader(filtered)
        for row in reader:
            dataset_id = (row.get("dataset") or "").strip()
            if not dataset_id:
                continue
            class_name = dataset_id.upper()
            if class_name not in wanted:
                continue
            rows[class_name] = row

    return rows


def _format_metadata_section(row: Mapping[str, str] | None, class_name: str) -> str:
    if not row:
        return (
            "Metadata for this dataset is not available in ``dataset_summary.csv``.\n"
        )

    def _clean(key: str, *, default: str | None = "") -> str:
        value = row.get(key, "")
        text = str(value).strip()
        if not text and default is not None:
            return default
        return text

    dataset_id = _clean("dataset", default="").lower() or class_name.lower()
    dataset_upper = dataset_id.upper()

    summary_bits: list[str] = []
    modality = _clean("modality of exp")
    if modality:
        summary_bits.append(f"Modality: {modality}")
    exp_type = _clean("type of exp")
    if exp_type:
        summary_bits.append(f"Type: {exp_type}")
    subject_type = _clean("Type Subject")
    if subject_type:
        summary_bits.append(f"Subjects: {subject_type}")

    lines = [f"- **Dataset ID:** ``{dataset_upper}``"]
    if summary_bits:
        lines.append(f"- **Summary:** {' | '.join(summary_bits)}")

    def _add_numeric(label: str, key: str) -> None:
        value = _clean(key)
        if value:
            lines.append(f"- **{label}:** {value}")

    _add_numeric("Number of Subjects", "n_subjects")
    _add_numeric("Number of Recordings", "n_records")
    _add_numeric("Number of Tasks", "n_tasks")
    _add_numeric("Number of Channels", "nchans_set")
    _add_numeric("Sampling Frequencies", "sampling_freqs")
    _add_numeric("Total Duration (hours)", "duration_hours_total")
    _add_numeric("Dataset Size", "size")

    nemar_url = f"https://nemar.org/dataexplorer/detail?dataset_id={dataset_id}"
    openneuro_url = f"https://openneuro.org/datasets/{dataset_id}"
    lines.append(f"- **OpenNeuro:** `{dataset_id} <{openneuro_url}>`__")
    lines.append(f"- **NeMAR:** `{dataset_id} <{nemar_url}>`__")

    table_row = {
        "dataset": dataset_id,
        "#Subj": _clean("n_subjects"),
        "#Chan": _clean("nchans_set"),
        "#Classes": _clean("n_tasks"),
        "Freq(Hz)": _clean("sampling_freqs"),
        "Duration(H)": _clean("duration_hours_total"),
        "Size": _clean("size"),
    }

    table = tabulate([table_row], headers="keys", tablefmt="rst", showindex=False)

    bullet_text = "\n".join(lines)
    return f"{bullet_text}\n\n{table}\n"


def _format_usage_section(class_name: str) -> str:
    return (
        ".. code-block:: python\n\n"
        f"   from eegdash.dataset import {class_name}\n\n"
        f'   dataset = {class_name}(cache_dir="./data")\n\n'
        '   print(f"Number of recordings: {len(dataset)}")\n\n'
        "   if len(dataset):\n"
        "       recording = dataset[0]\n"
        "       raw = recording.load()\n"
        "       print(f\"Sampling rate: {raw.info['sfreq']} Hz\")\n"
        '       print(f"Channels: {len(raw.ch_names)}")\n'
    )


def _format_see_also_section(dataset_id: str) -> str:
    dataset_lower = dataset_id.lower()
    nemar_url = f"https://nemar.org/dataexplorer/detail?dataset_id={dataset_lower}"
    openneuro_url = f"https://openneuro.org/datasets/{dataset_lower}"
    return "\n".join(
        [
            "* :class:`eegdash.dataset.EEGDashDataset`",
            "* :mod:`eegdash.dataset`",
            f"* `OpenNeuro dataset page <{openneuro_url}>`__",
            f"* `NeMAR dataset page <{nemar_url}>`__",
        ]
    )


def _cleanup_stale_dataset_pages(dataset_dir: Path, expected: set[Path]) -> None:
    for path in dataset_dir.glob("eegdash.dataset.DS*.rst"):
        if path in expected:
            continue
        try:
            if not path.read_text(encoding="utf-8").startswith(AUTOGEN_NOTICE):
                continue
        except OSError:
            continue
        path.unlink()


def _generate_dataset_docs(app) -> None:
    dataset_dir = Path(app.srcdir) / "api" / "dataset"
    dataset_dir.mkdir(parents=True, exist_ok=True)

    dataset_names = _iter_dataset_classes()
    dataset_rows = _load_dataset_rows(dataset_names)
    toctree_entries = _render_toctree_entries(dataset_names)
    experiment_rows = _render_experiment_rows(_load_experiment_counts(dataset_names))
    index_content = DATASET_INDEX_TEMPLATE.format(
        notice=AUTOGEN_NOTICE,
        dataset_count=len(dataset_names),
        experiment_rows=experiment_rows,
        toctree_entries=toctree_entries,
    )

    index_path = dataset_dir / "api_dataset.rst"
    if _write_if_changed(index_path, index_content):
        LOGGER.info("[dataset-docs] Updated %s", index_path.relative_to(app.srcdir))

    base_content = BASE_DATASET_TEMPLATE.format(notice=AUTOGEN_NOTICE)
    base_path = dataset_dir / "eegdash.dataset.EEGChallengeDataset.rst"
    if _write_if_changed(base_path, base_content):
        LOGGER.info("[dataset-docs] Updated %s", base_path.relative_to(app.srcdir))

    generated_paths: set[Path] = set()
    for name in dataset_names:
        title = f"eegdash.dataset.{name}"
        row = dataset_rows.get(name)
        dataset_id = ((row.get("dataset") if row else "") or name.lower()).strip()
        dataset_id = dataset_id or name.lower()
        page_content = DATASET_PAGE_TEMPLATE.format(
            notice=AUTOGEN_NOTICE,
            title=title,
            underline="=" * len(title),
            class_name=name,
            metadata_section=_format_metadata_section(row, name),
            usage_section=_format_usage_section(name),
            see_also_section=_format_see_also_section(dataset_id),
        )
        page_path = dataset_dir / f"eegdash.dataset.{name}.rst"
        if _write_if_changed(page_path, page_content):
            LOGGER.info("[dataset-docs] Updated %s", page_path.relative_to(app.srcdir))
        generated_paths.add(page_path)

    _cleanup_stale_dataset_pages(dataset_dir, generated_paths)

    # Remove legacy pages that used the short filename convention
    for legacy in dataset_dir.glob("DS*.rst"):
        try:
            legacy.unlink()
        except OSError:
            continue


def setup(app):
    """Create the back-references directory if it doesn't exist."""
    backreferences_dir = os.path.join(
        app.srcdir, sphinx_gallery_conf["backreferences_dir"]
    )
    if not os.path.exists(backreferences_dir):
        os.makedirs(backreferences_dir)

    app.connect("builder-inited", _generate_dataset_docs)


# Configure sitemap URL format (omit .html where possible)
sitemap_url_scheme = "{link}"

# Copy button configuration: strip common interactive prompts when copying
copybutton_prompt_text = r">>> |\\$ |# "
copybutton_prompt_is_regexp = True



import shutil
from pathlib import Path

import numpy as np
import pytest
import torch
from sklearn.model_selection import train_test_split
from sklearn.utils import check_random_state
from skorch.helper import SliceDataset
from torch.nn import functional as F
from torch.utils.data import DataLoader, TensorDataset

from braindecode.datautil import load_concat_dataset
from braindecode.models import ShallowFBCSPNet
from braindecode.preprocessing import (
    Preprocessor,
    create_windows_from_events,
    preprocess,
)
from braindecode.util import set_random_seeds
from eegdash import EEGDashDataset
from eegdash.hbn.preprocessing import hbn_ec_ec_reannotation
from eegdash.logging import logger

seed = 42
set_random_seeds(seed, cuda=False)
random_state = check_random_state(seed)


@pytest.fixture(scope="module")
def eeg_dash_dataset(cache_dir: Path):
    """Fixture to create an instance of EEGDashDataset."""
    return EEGDashDataset(
        query={
            "dataset": "ds005514",
            "task": "RestingState",
            "subject": "NDARDB033FW5",
        },
        cache_dir=cache_dir,
    )


@pytest.fixture(scope="module")
def preprocess_instance(eeg_dash_dataset, cache_dir: Path):
    """Fixture to create an instance of EEGDashDataset with preprocessing."""
    selected_channels = [
        "E22",
        "E9",
        "E33",
        "E24",
        "E11",
        "E124",
        "E122",
        "E29",
        "E6",
        "E111",
        "E45",
        "E36",
        "E104",
        "E108",
        "E42",
        "E55",
        "E93",
        "E58",
        "E52",
        "E62",
        "E92",
        "E96",
        "E70",
        "Cz",
    ]
    pre_processed_dir = cache_dir / "preprocessed"
    pre_processed_dir.mkdir(parents=True, exist_ok=True)
    try:
        eeg_dash_dataset = load_concat_dataset(
            pre_processed_dir,
            preload=True,
        )
        return eeg_dash_dataset

    except Exception as e:
        logger.warning(f"Failed to load dataset creating a new instance: {e}. ")
        if pre_processed_dir.exists():
            # folder with issue, erasing and creating again
            shutil.rmtree(pre_processed_dir)
            pre_processed_dir.mkdir(parents=True, exist_ok=True)

        preprocessors = [
            hbn_ec_ec_reannotation(),
            Preprocessor(
                "pick_channels",
                ch_names=selected_channels,
            ),
            Preprocessor("resample", sfreq=128),
            Preprocessor("filter", l_freq=1, h_freq=55),
        ]

        eeg_dash_dataset = preprocess(
            eeg_dash_dataset, preprocessors, n_jobs=-1, save_dir=pre_processed_dir
        )

        return eeg_dash_dataset


@pytest.fixture(scope="module")
def windows_ds(preprocess_instance):
    """Fixture to create windows from the preprocessed EEG dataset."""
    windows = create_windows_from_events(
        preprocess_instance,
        trial_start_offset_samples=0,
        trial_stop_offset_samples=256,
        preload=True,
    )
    return windows


def normalize_data(x):
    mean = x.mean(dim=2, keepdim=True)
    std = x.std(dim=2, keepdim=True) + 1e-7  # add small epsilon for numerical stability
    x = (x - mean) / std
    x = x.to(dtype=torch.float32)
    return x


def test_complete_train(windows_ds):
    """Test the complete training process with the EEG dataset."""
    label_eye_open_closed = np.array(
        SliceDataset(windows_ds, idx=1)
    ).T  # Extract labels (eyes open/closed)

    # train-test split
    train_idx, valid_idx = train_test_split(
        range(len(windows_ds)),
        test_size=0.2,
        stratify=label_eye_open_closed,
        random_state=random_state,
    )

    # Convert the data to tensors
    X_train = SliceDataset(windows_ds, idx=0, indices=train_idx)
    # Convert list of arrays to torch
    X_train = torch.FloatTensor(np.array(X_train))

    X_valid = SliceDataset(windows_ds, idx=0, indices=valid_idx)
    # Convert list of arrays to torch
    X_valid = torch.FloatTensor(np.array(X_valid))
    # Convert targets to tensor
    y_train = torch.LongTensor(label_eye_open_closed[train_idx])
    y_valid = torch.LongTensor(label_eye_open_closed[valid_idx])

    dataset_train = TensorDataset(X_train, y_train)
    dataset_valid = TensorDataset(X_valid, y_valid)

    # Create data loaders for training and valid (batch size 10)
    train_loader = DataLoader(dataset_train, batch_size=10, shuffle=True)
    val_loader = DataLoader(dataset_valid, batch_size=10, shuffle=False)

    logger.info(
        f"Shape of data {X_train.shape} number of samples - Train: {len(train_loader)}, Test: {len(val_loader)}"
    )
    logger.info(
        f"Eyes-Open/Eyes-Closed balance, train: {np.mean(label_eye_open_closed[train_idx]):.2f}, test: {np.mean(label_eye_open_closed[valid_idx]):.2f}"
    )

    model = ShallowFBCSPNet(n_chans=24, n_outputs=2, n_times=256)

    optimizer = torch.optim.AdamW(model.parameters(), lr=0.002, weight_decay=0.001)
    scheduler = torch.optim.lr_scheduler.ExponentialLR(optimizer, gamma=1)

    device = torch.device("cpu")
    model = model.to(device=device)  # move the model parameters to CPU/GPU
    epochs = 3

    for e in range(epochs):
        # training
        correct_train = 0
        for _, (x, y) in enumerate(train_loader):
            model.train()  # put model to training mode
            scores = model(normalize_data(x))
            _, preds = scores.max(1)
            correct_train += (preds == y).sum() / len(dataset_train)

            loss = F.cross_entropy(scores, y)
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()
            scheduler.step()

        # Validation
        correct_val = 0

        for t, (x, y) in enumerate(val_loader):
            model.eval()  # put model to testing mode
            scores = model(normalize_data(x))
            y = y.to(device=device, dtype=torch.long)
            _, preds = scores.max(1)
            correct_val += (preds == y).sum() / len(dataset_valid)

        logger.info(
            f"Epoch {e}, Train accuracy: {correct_train:.2f}, \
            Valid accuracy: {correct_val:.2f}"
        )



from pathlib import Path

import numpy as np
import pytest

from eegdash.dataset.dataset import EEGChallengeDataset


@pytest.fixture(scope="session")
def warmed_mongo(cache_dir: Path):
    """Skip tests gracefully if Mongo is not reachable."""
    try:
        # Lazy import to avoid circulars; constructing EEGChallengeDataset will touch DB
        _ = EEGChallengeDataset(release="R5", mini=True, cache_dir=cache_dir)
    except Exception:
        pytest.skip("Mongo not reachable")


def test_minirelease_vs_full_counts_and_subjects(warmed_mongo, cache_dir: Path):
    """Mini release should have fewer files and (typically) fewer subjects than full release."""
    release = "R5"

    ds_mini = EEGChallengeDataset(release=release, mini=True, cache_dir=cache_dir)
    ds_full = EEGChallengeDataset(release=release, mini=False, cache_dir=cache_dir)

    # File count: mini must be strictly smaller than full
    assert len(ds_mini.datasets) < len(ds_full.datasets)

    # Subject cardinality: mini should be strictly less than full, and > 0
    subj_mini = ds_mini.description["subject"].nunique()
    subj_full = ds_full.description["subject"].nunique()
    assert subj_mini > 0
    assert subj_mini < subj_full


def test_minirelease_subject_raw_equivalence(warmed_mongo, cache_dir: Path):
    """For a subject present in the mini set, loading that subject in mini vs full yields identical raw data."""
    release = "R5"

    # Pick a concrete subject from the mini set to avoid guessing
    ds_mini_all = EEGChallengeDataset(release=release, mini=True, cache_dir=cache_dir)
    assert len(ds_mini_all.datasets) > 0
    subject = ds_mini_all.description["subject"].iloc[0]

    ds_mini = EEGChallengeDataset(
        release=release,
        mini=True,
        cache_dir=cache_dir,
    )
    ds_full = EEGChallengeDataset(
        release=release, mini=False, cache_dir=cache_dir, subject=subject
    )

    assert len(ds_mini.datasets) > 0
    assert len(ds_full.datasets) > 0

    # Identify a common BIDS file (bidspath) present in both (bucket prefixes differ between mini/full)
    mini_paths = {d.record["bidspath"] for d in ds_mini.datasets}
    full_paths = {d.record["bidspath"] for d in ds_full.datasets}
    intersection = mini_paths & full_paths
    assert intersection, "No common recordings found for the chosen subject"

    common_path = next(iter(intersection))
    mini_idx = next(
        i for i, d in enumerate(ds_mini.datasets) if d.record["bidspath"] == common_path
    )
    full_idx = next(
        i for i, d in enumerate(ds_full.datasets) if d.record["bidspath"] == common_path
    )

    raw_mini = ds_mini.datasets[mini_idx].raw
    raw_full = ds_full.datasets[full_idx].raw

    # Basic metadata equivalence
    assert raw_mini.info["sfreq"] == raw_full.info["sfreq"]
    assert raw_mini.info["nchan"] == raw_full.info["nchan"]
    assert raw_mini.ch_names == raw_full.ch_names

    # Compare a small data slice to ensure content equality (avoid loading entire arrays into memory)
    n_samples = min(1000, raw_mini.n_times, raw_full.n_times)
    assert n_samples > 0
    data_mini = raw_mini.get_data(picks=[0], start=0, stop=n_samples)
    data_full = raw_full.get_data(picks=[0], start=0, stop=n_samples)
    assert np.allclose(data_mini, data_full, rtol=1e-6, atol=0), (
        "Raw data mismatch between mini and full"
    )


def test_minirelease_consume_everything(warmed_mongo, cache_dir: Path):
    """Simply try to load all data in the mini release to catch any errors."""
    release = "R5"
    ds_mini = EEGChallengeDataset(release=release, mini=True, cache_dir=cache_dir)

    for dataset in ds_mini.datasets:
        raw = dataset.raw  # noqa: F841
        description = dataset.description  # noqa: F841
        assert raw is not None
        assert description is not None



"""Test for features module Python 3.10+ compatibility."""

import pytest


def test_import_features_module():
    """Test that the features module can be imported without syntax errors.

    This test ensures Python 3.10+ compatibility by verifying that:
    1. Type annotations with list[], type[], and | syntax work (via __future__ imports)
    2. No Python 3.11+ exclusive syntax is used (like *unpacking in subscripts)
    """
    try:
        import eegdash.features

        assert eegdash.features is not None
    except SyntaxError as e:
        pytest.fail(f"SyntaxError when importing eegdash.features: {e}")
    except ImportError as e:
        pytest.fail(f"ImportError when importing eegdash.features: {e}")


def test_import_features_submodules():
    """Test that all features submodules can be imported."""
    submodules = [
        "eegdash.features.inspect",
        "eegdash.features.extractors",
        "eegdash.features.serialization",
        "eegdash.features.datasets",
        "eegdash.features.decorators",
        "eegdash.features.feature_bank",
        "eegdash.features.feature_bank.complexity",
        "eegdash.features.feature_bank.dimensionality",
        "eegdash.features.feature_bank.signal",
        "eegdash.features.feature_bank.spectral",
        "eegdash.features.feature_bank.connectivity",
        "eegdash.features.feature_bank.csp",
    ]

    for module_name in submodules:
        try:
            __import__(module_name)
        except SyntaxError as e:
            pytest.fail(f"SyntaxError when importing {module_name}: {e}")
        except ImportError:
            # Some imports might fail due to missing dependencies, that's ok
            # We only care about SyntaxError
            pass


def test_features_basic_functionality():
    """Test basic features module functionality."""
    from eegdash.features import (
        get_all_feature_extractors,
        get_all_feature_kinds,
        get_all_features,
    )

    # These should return lists without errors
    features = get_all_features()
    assert isinstance(features, list)

    extractors = get_all_feature_extractors()
    assert isinstance(extractors, list)

    kinds = get_all_feature_kinds()
    assert isinstance(kinds, list)



from __future__ import annotations

import inspect
from collections.abc import Callable

from . import extractors, feature_bank
from .extractors import FeatureExtractor, MultivariateFeature, _get_underlying_func


def get_feature_predecessors(feature_or_extractor: Callable) -> list:
    """Get the dependency hierarchy for a feature or feature extractor.

    This function recursively traverses the `parent_extractor_type` attribute
    of a feature or extractor to build a list representing its dependency
    lineage.

    Parameters
    ----------
    feature_or_extractor : callable
        The feature function or :class:`FeatureExtractor` class to inspect.

    Returns
    -------
    list
        A nested list representing the dependency tree. For a simple linear
        chain, this will be a flat list from the specific feature up to the
        base `FeatureExtractor`. For multiple dependencies, it will contain
        tuples of sub-dependencies.

    """
    current = _get_underlying_func(feature_or_extractor)
    if current is FeatureExtractor:
        return [current]
    predecessor = getattr(current, "parent_extractor_type", [FeatureExtractor])
    if len(predecessor) == 1:
        return [current, *get_feature_predecessors(predecessor[0])]
    else:
        predecessors = [get_feature_predecessors(pred) for pred in predecessor]
        for i in range(len(predecessors)):
            if isinstance(predecessors[i], list) and len(predecessors[i]) == 1:
                predecessors[i] = predecessors[i][0]
        return [current, tuple(predecessors)]


def get_feature_kind(feature: Callable) -> MultivariateFeature:
    """Get the 'kind' of a feature function.

    The feature kind (e.g., univariate, bivariate) is typically attached by a
    decorator.

    Parameters
    ----------
    feature : callable
        The feature function to inspect.

    Returns
    -------
    MultivariateFeature
        An instance of the feature kind (e.g., `UnivariateFeature()`).

    """
    return _get_underlying_func(feature).feature_kind


def get_all_features() -> list[tuple[str, Callable]]:
    """Get a list of all available feature functions.

    Scans the `eegdash.features.feature_bank` module for functions that have
    been decorated to have a `feature_kind` attribute.

    Returns
    -------
    list[tuple[str, callable]]
        A list of (name, function) tuples for all discovered features.

    """

    def isfeature(x):
        return hasattr(_get_underlying_func(x), "feature_kind")

    return inspect.getmembers(feature_bank, isfeature)


def get_all_feature_extractors() -> list[tuple[str, type[FeatureExtractor]]]:
    """Get a list of all available `FeatureExtractor` classes.

    Scans the `eegdash.features.feature_bank` module for all classes that
    subclass :class:`~eegdash.features.extractors.FeatureExtractor`.

    Returns
    -------
    list[tuple[str, type[FeatureExtractor]]]
        A list of (name, class) tuples for all discovered feature extractors,
        including the base `FeatureExtractor` itself.

    """

    def isfeatureextractor(x):
        return inspect.isclass(x) and issubclass(x, FeatureExtractor)

    return [
        ("FeatureExtractor", FeatureExtractor),
        *inspect.getmembers(feature_bank, isfeatureextractor),
    ]


def get_all_feature_kinds() -> list[tuple[str, type[MultivariateFeature]]]:
    """Get a list of all available feature 'kind' classes.

    Scans the `eegdash.features.extractors` module for all classes that
    subclass :class:`~eegdash.features.extractors.MultivariateFeature`.

    Returns
    -------
    list[tuple[str, type[MultivariateFeature]]]
        A list of (name, class) tuples for all discovered feature kinds.

    """

    def isfeaturekind(x):
        return inspect.isclass(x) and issubclass(x, MultivariateFeature)

    return inspect.getmembers(extractors, isfeaturekind)



import time
from pathlib import Path

import pytest

from eegdash.api import EEGDash
from eegdash.dataset import EEGChallengeDataset

RELEASES = ["R1", "R2", "R3", "R4", "R5", "R6", "R7", "R8", "R9", "R10", "R11"]
FILES_PER_RELEASE = [1342, 1405, 1812, 3342, 3326, 1227, 3100, 2320, 2885, 2516, 3397]

RELEASE_FILES = list(zip(RELEASES, FILES_PER_RELEASE))


def _load_release(release, cache_dir: Path):
    ds = EEGChallengeDataset(release=release, mini=False, cache_dir=cache_dir)
    getattr(ds, "description", None)
    return ds


def test_eeg_challenge_dataset_initialization(cache_dir: Path):
    """Test the initialization of EEGChallengeDataset."""
    dataset = EEGChallengeDataset(release="R5", mini=False, cache_dir=cache_dir)

    release = "R5"
    expected_bucket_prefix = f"s3://nmdatasets/NeurIPS25/{release}_L100_bdf"
    assert dataset.s3_bucket == expected_bucket_prefix, (
        f"Unexpected s3_bucket: {dataset.s3_bucket} (expected {expected_bucket_prefix})"
    )

    # Expected components (kept explicit for readability & easier future edits)
    expected_dataset = "ds005509"
    expected_subject = "sub-NDARAC350XUM"
    expected_task = "DespicableMe"
    expected_suffix = (
        f"{expected_dataset}/{expected_subject}/eeg/"
        f"{expected_subject}_task-{expected_task}_eeg.set"
    )

    expected_full_path = f"{dataset.s3_bucket}/{expected_suffix}"
    first_file = dataset.datasets[0].s3file

    assert first_file == expected_full_path, (
        "Mismatch in first dataset s3 file path.\n"
        f"Got     : {first_file}\n"
        f"Expected: {expected_full_path}"
    )


@pytest.mark.parametrize("release, number_files", RELEASE_FILES)
def test_eeg_challenge_dataset_amount_files(release, number_files, cache_dir: Path):
    dataset = EEGChallengeDataset(release=release, mini=False, cache_dir=cache_dir)
    assert len(dataset.datasets) == number_files


@pytest.mark.parametrize("release", RELEASES)
def test_mongodb_load_benchmark(benchmark, release, cache_dir: Path):
    # Group makes the report nicer when comparing releases
    benchmark.group = "EEGChallengeDataset.load"

    result = benchmark.pedantic(
        _load_release,
        args=(release, cache_dir),
        iterations=1,  # I/O-bound → 1 iteration per round
        rounds=5,  # take min/median across several cold-ish runs
        warmup_rounds=1,  # do one warmup round
    )

    assert result is not None


@pytest.mark.parametrize("release", RELEASES)
def test_mongodb_load_under_sometime(release, cache_dir: Path):
    start_time = time.perf_counter()
    _ = EEGChallengeDataset(release=release, cache_dir=cache_dir)
    duration = time.perf_counter() - start_time
    assert duration < 30, f"{release} took {duration:.2f}s"


@pytest.mark.parametrize("mini", [True, False])
@pytest.mark.parametrize("release", RELEASES)
def test_consuming_one_raw(release, mini, cache_dir: Path):
    print(f"Testing release {release} mini={mini} with cache dir {cache_dir}")
    dataset_obj = EEGChallengeDataset(
        release=release,
        task="RestingState",
        cache_dir=cache_dir,
        mini=mini,
    )
    raw = dataset_obj.datasets[0].raw
    assert raw is not None


@pytest.mark.parametrize("eeg_dash_instance", [None, EEGDash()])
def test_eeg_dash_integration(
    eeg_dash_instance, cache_dir: Path, release="R5", mini=True
):
    dataset_obj = EEGChallengeDataset(
        release=release,
        task="RestingState",
        cache_dir=cache_dir,
        mini=mini,
        eeg_dash_instance=eeg_dash_instance,
    )
    raw = dataset_obj.datasets[0].raw
    assert raw is not None



from pathlib import Path

from eegdash.api import EEGDashDataset
from eegdash.dataset.dataset import EEGChallengeDataset


def test_challenge_dataset_passes_task_and_dataset_filters(cache_dir: Path):
    ds = EEGChallengeDataset(
        release="R5",
        cache_dir=cache_dir,
        mini=False,
        task="RestingState",
    )

    assert len(ds.datasets) == 329
    assert ds.datasets[0].record["dataset"] == "ds005509"


def test_eegdashdataset_ignores_empty_query_when_kwargs_present(cache_dir: Path):
    _ = EEGDashDataset(
        query={},
        cache_dir=cache_dir,
        dataset="ds005509",
        task="RestingState",
    )


def test_challenge_dataset_task_list_propagation(cache_dir: Path):
    tasks = ["RestingState", "DespicableMe"]

    _ = EEGChallengeDataset(
        release="R5",
        cache_dir=cache_dir,
        mini=False,
        task=tasks,
    )


def test_eegdashdataset_allows_raw_query_and_kwargs(cache_dir: Path):
    _ = EEGDashDataset(
        query={"subject": {"$in": ["NDARAU708TL8", "NDARAP785CTE"]}},
        cache_dir=cache_dir,
        dataset="ds005509",
        task="RestingState",
    )


def test_challenge_dataset_mini_populates_subjects(cache_dir: Path):
    _ = EEGChallengeDataset(
        release="R5",
        cache_dir=cache_dir,
        mini=True,
        task="RestingState",
    )



from pathlib import Path

from eegdash.bids_eeg_metadata import load_eeg_attrs_from_bids_file


class _DummyBIDSDataset:
    def __init__(self, root: Path, bids_file: Path):
        self.bidsdir = root
        self.dataset = root.name
        self.files = [str(bids_file)]

    def _get_relative_bidspath(self, filename: str | Path) -> str:
        return str(Path(filename).relative_to(self.bidsdir.parent))

    get_relative_bidspath = _get_relative_bidspath

    def subject_participant_tsv(self, bids_file):  # noqa: ARG002 - unused in tests
        return None

    def eeg_json(self, bids_file):  # noqa: ARG002 - unused in tests
        return None

    def get_bids_metadata_files(self, bids_file, extension):  # noqa: ARG002 - unused
        return []

    def get_bids_file_attribute(self, attribute: str, bids_file: str):
        defaults = {
            "subject": "01",
            "session": "01",
            "task": "test",
            "run": "01",
            "modality": "eeg",
            "sfreq": 100,
            "nchans": 1,
            "ntimes": 1,
        }
        return defaults.get(attribute)


# --- tests -----------------------------------------------------------------


def test_sidecar_dependencies_filtered(tmp_path: Path):
    root = tmp_path / "ds000001"
    eeg_dir = root / "sub-01" / "ses-01" / "eeg"
    eeg_dir.mkdir(parents=True)
    base = "sub-01_ses-01_task-test_run-01_eeg"
    vhdr = eeg_dir / f"{base}.vhdr"
    vhdr.write_text("")
    eeg = eeg_dir / f"{base}.eeg"
    eeg.write_text("")
    vmrk = eeg_dir / f"{base}.vmrk"
    vmrk.write_text("")

    bids_ds = _DummyBIDSDataset(root, vhdr)
    attrs = load_eeg_attrs_from_bids_file(bids_ds, str(vhdr))

    expected = {
        bids_ds._get_relative_bidspath(eeg),
        bids_ds._get_relative_bidspath(vmrk),
    }
    assert expected.issubset(set(attrs["bidsdependencies"]))
    assert not any(
        dep.endswith(ext)
        for ext in [".dat", ".raw", ".vhdr"]
        for dep in attrs["bidsdependencies"]
    )



# Authors: The EEGDash contributors.
# License: GNU General Public License
# Copyright the EEGDash contributors.

"""General utility functions for EEGDash.

This module contains miscellaneous utility functions used across the EEGDash package,
including MongoDB client initialization and configuration helpers.
"""

from mne.utils import get_config, set_config, use_log_level


def _init_mongo_client() -> None:
    """Initialize the default MongoDB connection URI in the MNE config.

    This function checks if the ``EEGDASH_DB_URI`` is already set in the
    MNE-Python configuration. If it is not set, this function sets it to the
    default public EEGDash MongoDB Atlas cluster URI.

    The operation is performed with MNE's logging level temporarily set to
    "ERROR" to suppress verbose output.

    Notes
    -----
    This is an internal helper function and is not intended for direct use
    by end-users.

    """
    with use_log_level("ERROR"):
        if get_config("EEGDASH_DB_URI") is None:
            set_config(
                "EEGDASH_DB_URI",
                "mongodb+srv://eegdash-user:mdzoMjQcHWTVnKDq@cluster0.vz35p.mongodb.net/?retryWrites=true&w=majority&appName=Cluster0",
                set_env=True,
            )


__all__ = ["_init_mongo_client"]
